import contextlib
import fnmatch
import glob
import json
import os
from functools import reduce
from typing import Optional, Set, List, Dict

import click
import yaml
from box import Box
from marshmallow import ValidationError

from montecarlodata.config import Config
from montecarlodata.errors import complain_and_abort, manage_errors
from montecarlodata.fs_utils import mkdirs
from montecarlodata.iac.client import MonteCarloConfigTemplateClient
from montecarlodata.iac.schemas import ProjectConfig
from montecarlodata.iac.utils import is_dbt_schema, has_montecarlo_property
from montecarlodata.settings import TARGET_DIRECTORY_NAME, PROJECT_CONFIG_FILENAME
from montecarlodata.utils import GqlWrapper


class MonteCarloConfigService:
    def __init__(self, config: Config, request_wrapper: Optional[GqlWrapper] = None, project_dir: str = None):
        request_wrapper = request_wrapper or GqlWrapper(mcd_id=config.mcd_id, mcd_token=config.mcd_token)
        self.client = MonteCarloConfigTemplateClient(request_wrapper)

        if not project_dir:
            self.project_dir = os.getcwd()
        else:
            self.project_dir = os.path.abspath(project_dir)
        self.init()
        self._abort_on_error = True

    def init(self):
        self.target_dir = self._create_target_directory()
        self.project_config = self._load_project_config()

    @manage_errors
    def apply(self, namespace: str, dry_run: bool = False, abort_on_error=True):
        """
        Compile configuration.
        Submit configuration via API
        """
        files, mc_config_dict, _ = self.compile()
        response = self.client.apply_config_template(
            namespace,
            mc_config_dict,
            resource=self.project_config.default_resource,
            dry_run=dry_run
        )

        if response.errors:
            click.echo()
            click.echo('Errors encountered when attempting to apply configuration:')

            system_error = response.errors.get('system_error', {})
            if system_error:
                click.echo(f' - {system_error}')
            else:
                validation_errors = response.errors.get('validation_errors', {})
                click.echo(f' - Validation error: {validation_errors}')

            click.echo()

            if abort_on_error:
                complain_and_abort("Errors encountered, exiting.")

        if response.resource_modifications:
            click.echo()
            click.echo('Modifications:')
            for resource_modification in response.resource_modifications:
                click.echo(f' - {"[DRY RUN] " if dry_run else ""}{resource_modification.type} - {resource_modification.description}')
            click.echo()

            if dry_run:
                click.echo('Dry run, none of the changes have been applied.')
                click.echo()
            else:
                click.echo('Changes successfully applied.')
                click.echo()
        else:
            click.echo()
            click.echo('No changes to configuration found, doing nothing.')
            click.echo()

        return response

    @manage_errors
    def compile(self, abort_on_error=True) -> Dict:
        """
        Gather monitor configuration YAML files from project.
        This may also include DBT schema files with meta properties
        """
        click.echo()
        click.echo("Gathering monitor configuration files.")
        files = self._gather_yaml_files(self.project_config)

        errors_by_file = {}
        mc_config_list: List[Box] = []

        files = sorted(list(files))
        for file in files:
            with open(file, 'r') as f:
                yaml_as_dict = Box(yaml.safe_load(f))

                # This file is a DBT schema file.
                # Let's attempt to parse out special embedded MC sections from meta:
                # See https://docs.getdbt.com/reference/resource-properties/meta
                if is_dbt_schema(yaml_as_dict):
                    for model in yaml_as_dict.models:
                        with contextlib.suppress(KeyError):
                            mc_config = model.meta.montecarlo
                            errors = self._validate_mc_config(mc_config)
                            if errors:
                                errors_by_file[file] = errors
                            mc_config_list.append(mc_config)

                            click.echo(f" - {file} - Embedded monitor configuration found.")

                # This file has a root montecarlo property
                # This can be in a standalone file, and also a DBT schema file
                if has_montecarlo_property(yaml_as_dict):
                    mc_config = yaml_as_dict.montecarlo
                    errors = self._validate_mc_config(mc_config)
                    if errors:
                        errors_by_file[file] = errors
                    mc_config_list.append(mc_config)
                    click.echo(f" - {file} - Monitor configuration found.")

        # Exit if there are errors
        if errors_by_file:
            click.echo()
            click.echo('Configuration validation errors:')
            for file, errors in errors_by_file.items():
                click.echo(f' - File: {file}')
                for error in errors:
                    click.echo(f'    - {error}')
            click.echo()
            if abort_on_error:
                complain_and_abort("Errors encountered, exiting.")

        if not mc_config_list:
            if abort_on_error:
                complain_and_abort("Sorry, we didn't find any YAML files containing Monte Carlo configuration.")

        # Merge configs into a single config
        compiled_mc_config = self._merge_mc_configs(mc_config_list)

        # Write configs to target directory. Useful for debugging.
        with open(os.path.join(self.target_dir, 'montecarlo_configuration.yml'), 'w') as f:
            f.write(compiled_mc_config.to_yaml())

        with open(os.path.join(self.target_dir, 'montecarlo_configuration.json'), 'w') as f:
            json.dump(compiled_mc_config, f, indent=4)

        return files, compiled_mc_config, errors_by_file

    def _create_target_directory(self) -> str:
        target_dir = os.path.join(self.project_dir, TARGET_DIRECTORY_NAME)
        mkdirs(target_dir)
        return target_dir

    def _load_project_config(self) -> ProjectConfig:
        contents = None
        try:
            with open(os.path.join(self.project_dir, PROJECT_CONFIG_FILENAME), 'r') as f:
                contents = f.read()
        except FileNotFoundError:
            complain_and_abort(
                f"Not a Monte Carlo project. Must define {PROJECT_CONFIG_FILENAME} in current working directory.")

        project_config_as_dict = yaml.safe_load(contents)
        try:
            s = ProjectConfig.schema()
            return s.load(project_config_as_dict)
        except ValidationError as e:
            complain_and_abort(f"Encountered a validation problem in {PROJECT_CONFIG_FILENAME}: {e.messages}")

    def _merge_mc_configs(self, mc_config_list: List[Box]) -> Box:
        """
        Merge multiple mc configs into a single one.
        Nested dicts will be merged
        Nested lists will be concatenated
        """
        def merge_configs(x: Box, y: Box):
            x.merge_update(y, box_merge_lists='extend')
            return x

        compiled_mc_config = reduce(merge_configs, mc_config_list)
        compiled_mc_config.pop('box_merge_lists', None)  # Strangly enough, Box adds this add'l key. Remove it

        return compiled_mc_config

    def _validate_mc_config(self, mc_config_dict: Box) -> List[str]:
        """
        Perform simple validation.
        More comprehensive validation is performed server-side
        """
        expected_keys = [
            'field_health',
            'dimension_tracking',
            'json_schema',
            'custom_sql',
            'freshness',
            'volume'
        ]

        errors = []
        for key in expected_keys:
            with contextlib.suppress(KeyError):
                if key in mc_config_dict and not isinstance(mc_config_dict[key], list):
                    errors.append(f'"{key}" property should be a list.')
        return errors

    def _gather_yaml_files(self, project_config: ProjectConfig) -> Set[str]:
        files = set()
        for pattern in project_config.include_file_patterns:
            for fn in glob.glob(os.path.join(self.project_dir, pattern), recursive=True):
                files.add(fn)
        for pattern in project_config.exclude_file_patterns:
            excluded = set(fnmatch.filter(files, os.path.join(self.project_dir, pattern)))
            files = files - excluded
        return files

    @manage_errors
    def delete(self, namespace: str, dry_run: bool = False, abort_on_error=True):
        """
        Delete all monitors in namespace
        """
        response = self.client.delete_config_template(
            namespace,
            dry_run=dry_run
        )

        click.echo(f'\n{"[DRY RUN] " if dry_run else ""}Deleted {response.num_deleted} resources in namespace={namespace}\n')

        return response




from typing import Optional, List, Tuple, Iterator

import click
from tabulate import tabulate

from montecarlodata.common.common import is_overlap
from montecarlodata.common.data import AWSArn, DcResourceProperties
from montecarlodata.common.resources import CloudResourceService
from montecarlodata.errors import complain_and_abort, manage_errors
from montecarlodata.integrations.onboarding.fields import REDSHIFT_CONNECTION_TYPE


class NetworkDiscoveryService:
    MCD_NETWORK_REC_RESOURCE_TYPE_MAP = {
        REDSHIFT_CONNECTION_TYPE: 'rs_network_recommender'
    }

    def __init__(self, cloud_resource_service: Optional[CloudResourceService] = None, *args, **kwargs):
        self._abort_on_error = True  # Used in decorator
        self._cloud_resource_service = cloud_resource_service or CloudResourceService(*args, **kwargs)

    @manage_errors
    def recommend_network_dispatcher(self, *, resource_type: str, dc_id: Optional[str] = None, **kwargs) -> None:
        """
        Make a network connection recommendation by resource to help simplify steps required to onboard.
        """
        try:
            dispatch = self.MCD_NETWORK_REC_RESOURCE_TYPE_MAP[resource_type]
        except KeyError:
            complain_and_abort(f'Unsupported resource type - \'{resource_type}\'.')
        else:
            collector = self._cloud_resource_service.get_and_validate_active_collector(dc_id=dc_id)
            dc_resource_props = self._cloud_resource_service.get_dc_resource_props(collector_props=collector, **kwargs)

            getattr(self, dispatch)(dc_resource_props=dc_resource_props, **kwargs)

    def rs_network_recommender(self, *, dc_resource_props: DcResourceProperties, resource_identifier: str, **_) -> None:
        """
        Alpha recommender for how to connect the DC with RS.

        Connections can be made with either IP filtering or peering. Peering has the following scenarios -
            1. Deployed without a VPC
            2. Overlapping CIDR ranges
            3. Same account peering
            4. Multi account peering

        Multi-region is supported. Transit gateways are supported, but not considered by the recommender as MC currently
        pushes peering.

        https://docs.getmontecarlo.com/docs/aws-networking-troubleshoot-guide#after-data-collector-deployment
        """
        self._check_collector_validity(dc_resource_props)

        # Retrieve properties from the redshift cluster
        dc_resource_props.resource_props = dc_resource_props.resources_client.describe_redshift_cluster(cluster_id=resource_identifier)
        dc_resource_props.resource_arn = AWSArn(dc_resource_props.resource_props.cluster_namespace_arn)

        does_redshift_dc_share_an_account = dc_resource_props.resource_arn.account == dc_resource_props.collector_arn.account
        redshift_security_groups, redshift_in_vpc = self._get_rs_sg(dc_resource_props)

        redshift_vpc_range = dc_resource_props.resources_client.describe_vpc(dc_resource_props.resource_props.vpc_id).cidr_block
        collector_vpc_range = dc_resource_props.collector_props.parameters.vpc_cidr

        base_header = 'The redshift cluster \'{}\' is{}publicly accessible.'

        # Build recommendations based on the cluster's availability (public or not).
        if dc_resource_props.resource_props.publicly_accessible:
            header = base_header.format(resource_identifier, ' ')
            recommendations = self._get_ip_filtering_verbiage(dc_resource_props, redshift_security_groups)
        else:
            if not redshift_in_vpc:
                # Clusters deployed outside of a VPC are essentially a legacy feature and not supported.
                complain_and_abort('Recommender does not support non-public clusters deployed outside a VPC.')

            header = base_header.format(resource_identifier, ' not ')
            recommendation_intro = self._get_peering_intro_verbiage(does_redshift_dc_share_an_account)

            if is_overlap(redshift_vpc_range, collector_vpc_range):
                recommendations = [
                    *recommendation_intro,
                    self._get_overlap_verbiage(collector_vpc_range, redshift_vpc_range)
                ]  # This case is considered a recommendation and not an error as is a "supportable path".
            else:
                redshift_routes = self._get_rs_routes(dc_resource_props)

                recommendations = [
                    *recommendation_intro,
                    *self._get_rs_cross_account_verbiage(dc_resource_props, does_redshift_dc_share_an_account),
                    *self._get_rs_requester_verbiage(dc_resource_props, does_redshift_dc_share_an_account, redshift_vpc_range),
                    *self._get_rs_accepter_verbiage(dc_resource_props, redshift_security_groups, redshift_routes),
                    *self._get_extra_verbiage()
                ]

        self._echo_recommendation([*recommendations], header_col=header)

    @staticmethod
    def _check_collector_validity(dc_resource_props: DcResourceProperties, default_unset_val: Optional[str] = 'N/A') -> None:
        """
        Error on unsupported collectors (e.g. using an existing VPC).
        """
        if 'existing_vpc_id' in dc_resource_props.collector_props.parameters and \
                dc_resource_props.collector_props.parameters.existing_vpc_id != default_unset_val:
            complain_and_abort('Recommender does not support collectors deployed in customer managed VPCs.')

    @staticmethod
    def _get_rs_sg(dc_resource_props: DcResourceProperties, expected_status: Optional[str] = 'active') -> Tuple[List[str], bool]:
        """
        Parse the security groups associated with the RS cluster.

        Returns a list of SGs and a flag to indicate the cluster is deployed in a VPC (yes=True).
        """
        if all(prop in dc_resource_props.resource_props for prop in ['vpc_security_groups', 'cluster_security_groups']):
            if len(dc_resource_props.resource_props.vpc_security_groups) > 0 and len(
                    dc_resource_props.resource_props.cluster_security_groups) > 0:
                # This should not be possible as AWS does not support this case.
                complain_and_abort('Recommender does not support clusters with ambiguous hosting. '
                                   'Cluster cannot have security groups both inside and outside a VPC.')
            elif len(dc_resource_props.resource_props.vpc_security_groups) > 0:
                return [sg.vpc_security_group_id for sg in dc_resource_props.resource_props.vpc_security_groups
                        if sg.status.lower() == expected_status], True
            elif len(dc_resource_props.resource_props.cluster_security_groups) > 0:
                return [sg.cluster_security_group_name for sg in
                        dc_resource_props.resource_props.cluster_security_groups
                        if sg.status.lower() == expected_status], False
        return [], False

    @staticmethod
    def _get_rs_routes(dc_resource_props: DcResourceProperties) -> Iterator[str]:
        """
        Get routes for the cluster.
        """
        if 'cluster_subnet_group_name' not in dc_resource_props.resource_props:
            complain_and_abort('Recommender does not support clusters deployed outside a VPC.')

        rs_subnets = [subnet.subnet_identifier for subnet in
                      dc_resource_props.resources_client.describe_cluster_subnet_groups(
                          subnet_group=dc_resource_props.resource_props.cluster_subnet_group_name).subnets]

        for subnet in rs_subnets or []:
            route = dc_resource_props.resources_client.describe_routes(filter_vals=[subnet]).route_tables
            if len(route) == 0:
                route = dc_resource_props.resources_client.describe_routes(filter_vals=['true'],
                                                                           filter_key='association.main').route_tables
            yield route[0].route_table_id

    @staticmethod
    def _get_ip_filtering_verbiage(dc_resource_props: DcResourceProperties,
                                   redshift_security_groups: List[str]) -> List[str]:
        """
        Generate verbiage for IP filtering.
        """
        return [
            'IP filtering is recommended. See steps below.',
            '',
            f'Whitelist port \'{dc_resource_props.resource_props.endpoint.port}\' for '
            f'\'{dc_resource_props.collector_props.outputs.public_ip}/32\' in any of the following '
            f'redshift security groups - {redshift_security_groups}.',
            f'https://docs.getmontecarlo.com/docs/network-connectivity#ip-filtering'
        ]

    @staticmethod
    def _get_rs_cross_account_verbiage(dc_resource_props: DcResourceProperties,
                                       does_redshift_dc_share_an_account: bool) -> List[str]:
        """
        Generate verbiage for cross account stack stack. This is an optional step based on the accounts used.
        """
        if does_redshift_dc_share_an_account:
            return []
        return [
            '- Deploy the \'Create Peering Cross Account assumable role CloudFormation stack\' '
            'with the following values -',
            f'1. Data Collector AWS Account ID: {dc_resource_props.collector_arn.account}',
            'https://docs.getmontecarlo.com/docs/peering-templates#create-peering-cross-account-assumable-role-cloudformation-stack',
            ''
        ]

    @staticmethod
    def _get_rs_requester_verbiage(dc_resource_props: DcResourceProperties,
                                   does_redshift_dc_share_an_account: bool,
                                   redshift_vpc_range: str) -> List[str]:
        """
        Generate verbiage for requester stack.
        """
        cross_role_verbiage = []
        if not does_redshift_dc_share_an_account:
            # Only visible if the previous stack was necessary
            cross_role_verbiage = [
                '8. VPC peer role for cross AWS account connections: <PeeringRole from the output of the previous stack>'
            ]

        return [
            '- Deploy the  \'Create Requester CloudFormation stack\' with the following values -',
            f'1. Monte Carlo Data Collector VPC ID: \'{dc_resource_props.collector_props.outputs.vpc_id}\'',
            f'2. Monte Carlo Data Collector Security Group ID: \'{dc_resource_props.collector_props.outputs.security_group}\'',
            f'3. Monte Carlo Data Collector Route Table ID: \'{dc_resource_props.collector_props.outputs.private_route_table}\'',
            f'4. Warehouse/resource VPC ID: \'{dc_resource_props.resource_props.vpc_id}\'',
            f'5. Warehouse/resource AWS Account ID: \'{dc_resource_props.resource_arn.account}\'',
            f'6. Warehouse/resource AWS Region: \'{dc_resource_props.resource_arn.region}\'',
            f'7. Warehouse/resource CIDR Block: \'{redshift_vpc_range}\'',
            *cross_role_verbiage,
            'https://docs.getmontecarlo.com/docs/peering-templates#create-requester-cloudformation-stack',
            ''
        ]

    @staticmethod
    def _get_rs_accepter_verbiage(dc_resource_props: DcResourceProperties,
                                  redshift_security_groups: List[str], redshift_routes: Iterator[str]) -> List[str]:
        """
        Generate verbiage for accepter stack.
        """
        return [
            '- Deploy the  \'Create Accepter CloudFormation stack\' with the following values -',
            f'1. Monte Carlo Data Collector CIDR Block: \'{dc_resource_props.collector_props.parameters.vpc_cidr}\'',
            '2. Monte Carlo Data Collector Peering Connection: <PeeringConnection from the output of the previous stack>',
            f'3. Resource / Warehouse Security Group: {NetworkDiscoveryService._get_friendly_sg_verbiage(redshift_security_groups)}',
            *[f'{count + 3}. Resource / Warehouse Route Table #{count}: \'{route}\'' for count, route in
              enumerate(set(redshift_routes), start=1)],
            'https://docs.getmontecarlo.com/docs/peering-templates#create-accepter-cloudformation-stack',
            ''
        ]

    @staticmethod
    def _get_overlap_verbiage(collector_vpc_range: str, redshift_vpc_range: str) -> str:
        """
        Generate verbiage for overlaps.
        """
        return f'Collector range ({collector_vpc_range}) overlaps with resource ({redshift_vpc_range}). ' \
               f'Peering is not possible when peered VPCs use overlapping CIDR blocks. ' \
               f'Please redeploy the collector with a custom CIDR block then rerun recommender.'

    @staticmethod
    def _get_peering_intro_verbiage(does_redshift_dc_share_an_account: bool) -> List[str]:
        """
        Generate verbiage for peering header / intro.
        """
        section_count = 2 if does_redshift_dc_share_an_account else 3

        return [
            f'VPC Peering is recommended. See the outlined steps for each of the {section_count} sections below.',
            '',
            'The CloudFormation template can be found by following the link at the end of each section.',
            f'Please complete all {section_count} sections, and for any additional help, reach out to your Monte Carlo representative.',
            ''
        ]

    @staticmethod
    def _get_extra_verbiage() -> List[str]:
        """
        Generate any extra verbiage (i.e. not path specific).
        """
        return [
            'If the cluster\'s subnets use a non-default ACL the collector CIDR Block also likely need to be whitelisted.'
        ]

    @staticmethod
    def _get_friendly_sg_verbiage(redshift_security_groups: List[str]) -> str:
        """
        Get Human readable SG groups
        """
        return f'Any of the following - {",".join(redshift_security_groups)}' \
            if len(redshift_security_groups) > 1 else f'\'{redshift_security_groups[0]}\''

    @staticmethod
    def _echo_recommendation(recs: List[str],
                             header_col: Optional[str] = 'Recommendations',
                             table_format: Optional[str] = 'pipe') -> None:
        """
        Display recommendations in a fancy table.
        """
        click.echo(tabulate([[header_col]] + [[rec] for rec in recs], headers='firstrow', tablefmt=table_format))



import click

from montecarlodata.insights.data_insights import InsightsService
from montecarlodata.insights.fields import KEY_ASSETS_INSIGHT_NAME, FILE_SCHEME, S3_SCHEME, SCHEME_DELIM, \
    COVERAGE_OVERVIEW_INSIGHT_NAME, CLEANUP_SUGGESTIONS_INSIGHT_NAME, EVENTS_INSIGHT_NAME, READ_WRITE_INSIGHT_NAME, \
    INCIDENT_QUERY_INSIGHT_NAME, DETERIORATING_QUERY_INSIGHT_NAME, RULES_AND_SLIS_INSIGHT_NAME
from montecarlodata.tools import add_common_options

# Shared command verbiage
GET_VERBIAGE = f"""
\b\n
DESTINATION is the path where the insight will be written to.

Supported schemes:\n
    '{FILE_SCHEME}{SCHEME_DELIM}' - save insight locally.\n
    '{S3_SCHEME}{SCHEME_DELIM}' - save insight to S3.

Notice - Will overwrite a file if it exists in the path and create any missing directories or prefixes.
"""

# Options shared across commands
GET_OPTIONS = [
    click.argument('destination', required=True),
    click.option('--aws-profile', required=False, help=f'AWS profile. If not specified, the one in the Monte Carlo '
                                                       f'CLI profile is used when uploading to S3.'),
    click.option('--dry', required=False, is_flag=True, help='Echo temporary presigned URL for the insight and quit.')
]


@click.group(help='Aggregated insights on your tables.')
def insights():
    """
    Group for any insight related subcommands
    """
    pass


@insights.command(help='List insights details and availability.', name='list')
@click.pass_obj
def list_insights(ctx):
    InsightsService(config=ctx['config']).echo_insights()


@insights.command(help=f'Get key assets insight. {GET_VERBIAGE}')
@add_common_options(GET_OPTIONS)
@click.pass_obj
def get_key_assets(ctx, **kwargs):
    InsightsService(config=ctx['config']).get_insight(insight=KEY_ASSETS_INSIGHT_NAME, **kwargs)


@insights.command(help=f'Get coverage overview (monitors) insight. {GET_VERBIAGE}')
@add_common_options(GET_OPTIONS)
@click.pass_obj
def get_coverage_overview(ctx, **kwargs):
    InsightsService(config=ctx['config']).get_insight(insight=COVERAGE_OVERVIEW_INSIGHT_NAME, **kwargs)


@insights.command(help=f'Get cleanup suggestions insight. {GET_VERBIAGE}')
@add_common_options(GET_OPTIONS)
@click.pass_obj
def get_cleanup_suggestions(ctx, **kwargs):
    InsightsService(config=ctx['config']).get_insight(insight=CLEANUP_SUGGESTIONS_INSIGHT_NAME, **kwargs)


@insights.command(help=f'Get events insight. {GET_VERBIAGE}')
@add_common_options(GET_OPTIONS)
@click.pass_obj
def get_events(ctx, **kwargs):
    InsightsService(config=ctx['config']).get_insight(insight=EVENTS_INSIGHT_NAME, **kwargs)


@insights.command(help=f'Get table read/write activity insight. {GET_VERBIAGE}')
@add_common_options(GET_OPTIONS)
@click.pass_obj
def get_table_activity(ctx, **kwargs):
    InsightsService(config=ctx['config']).get_insight(insight=READ_WRITE_INSIGHT_NAME, **kwargs)


@insights.command(help=f'Get incident query changes insight. {GET_VERBIAGE}')
@add_common_options(GET_OPTIONS)
@click.pass_obj
def get_incident_queries(ctx, **kwargs):
    InsightsService(config=ctx['config']).get_insight(insight=INCIDENT_QUERY_INSIGHT_NAME, **kwargs)


@insights.command(help=f'Get deteriorating queries insight. {GET_VERBIAGE}')
@add_common_options(GET_OPTIONS)
@click.pass_obj
def get_deteriorating_queries(ctx, **kwargs):
    InsightsService(config=ctx['config']).get_insight(insight=DETERIORATING_QUERY_INSIGHT_NAME, **kwargs)


@insights.command(help=f'Get rule and SLI results insight. {GET_VERBIAGE}')
@add_common_options(GET_OPTIONS)
@click.pass_obj
def get_rule_results(ctx, **kwargs):
    InsightsService(config=ctx['config']).get_insight(insight=RULES_AND_SLIS_INSIGHT_NAME, **kwargs)



from montecarlodata.common.common import read_as_base64
from montecarlodata.config import Config
from montecarlodata.errors import manage_errors
from montecarlodata.integrations.onboarding.base import BaseOnboardingService
from montecarlodata.integrations.onboarding.fields import EXPECTED_GENERIC_DB_GQL_RESPONSE_FIELD, \
    REDSHIFT_CONNECTION_TYPE, REDSHIFT_WAREHOUSE_TYPE, SNOWFLAKE_CONNECTION_TYPE, SNOWFLAKE_WAREHOUSE_TYPE, \
    EXPECTED_SNOWFLAKE_GQL_RESPONSE_FIELD, BQ_WAREHOUSE_TYPE, EXPECTED_BQ_GQL_RESPONSE_FIELD, BQ_CONNECTION_TYPE
from montecarlodata.queries.onboarding import TEST_DATABASE_CRED_MUTATION, TEST_SNOWFLAKE_CRED_MUTATION, \
    TEST_BQ_CRED_MUTATION


class WarehouseOnboardingService(BaseOnboardingService):

    def __init__(self, config: Config, **kwargs):
        super().__init__(config, **kwargs)

    @manage_errors
    def onboard_redshift(self, **kwargs) -> None:
        """
        Onboard a redshift connection by validating and adding a connection.
        """
        kwargs['connectionType'] = REDSHIFT_CONNECTION_TYPE
        kwargs['warehouseType'] = REDSHIFT_WAREHOUSE_TYPE
        self.onboard(validation_query=TEST_DATABASE_CRED_MUTATION,
                     validation_response=EXPECTED_GENERIC_DB_GQL_RESPONSE_FIELD,
                     connection_type=REDSHIFT_CONNECTION_TYPE, **kwargs)

    @manage_errors
    def onboard_snowflake(self, **kwargs) -> None:
        """
        Onboard a snowflake connection by validating and adding a connection.
        """
        kwargs['warehouseType'] = SNOWFLAKE_WAREHOUSE_TYPE
        self.onboard(validation_query=TEST_SNOWFLAKE_CRED_MUTATION,
                     validation_response=EXPECTED_SNOWFLAKE_GQL_RESPONSE_FIELD,
                     connection_type=SNOWFLAKE_CONNECTION_TYPE, **kwargs)

    @manage_errors
    def onboard_bq(self, **kwargs) -> None:
        """
        Onboard a BigQuery connection by validating and adding a connection.

        Reads and encodes service file as base64.
        """
        kwargs['warehouseType'] = BQ_WAREHOUSE_TYPE
        kwargs['serviceJson'] = read_as_base64(kwargs.pop('ServiceFile')).decode('utf-8')
        self.onboard(validation_query=TEST_BQ_CRED_MUTATION,
                     validation_response=EXPECTED_BQ_GQL_RESPONSE_FIELD,
                     connection_type=BQ_CONNECTION_TYPE, **kwargs)



# Query responses
EXPECTED_ADD_CONNECTION_RESPONSE_FIELD = 'addConnection'
EXPECTED_GENERIC_DB_GQL_RESPONSE_FIELD = 'testDatabaseCredentials'
EXPECTED_HIVE_S3_GQL_RESPONSE_FIELD = 'testS3Credentials'
EXPECTED_HIVE_SQL_GQL_RESPONSE_FIELD = 'testHiveCredentials'
EXPECTED_PRESTO_SQL_GQL_RESPONSE_FIELD = 'testPrestoCredentials'
EXPECTED_PRESTO_S3_GQL_RESPONSE_FIELD = 'testS3Credentials'
EXPECTED_GLUE_GQL_RESPONSE_FIELD = 'testGlueCredentials'
EXPECTED_ATHENA_GQL_RESPONSE_FIELD = 'testAthenaCredentials'
EXPECTED_SPARK_GQL_RESPONSE_FIELD = 'testSparkCredentials'
EXPECTED_SNOWFLAKE_GQL_RESPONSE_FIELD = 'testSnowflakeCredentials'
EXPECTED_BQ_GQL_RESPONSE_FIELD = 'testBqCredentials'
EXPECTED_SELF_HOSTED_GQL_RESPONSE_FIELD = 'testSelfHostedCredentials'
EXPECTED_TOGGLE_EVENTS_GQL_RESPONSE_FIELD = 'toggleEventConfig'
EXPECTED_UPDATE_CREDENTIALS_RESPONSE_FIELD = 'updateCredentials'
EXPECTED_REMOVE_CONNECTION_RESPONSE_FIELD = 'removeConnection'
EXPECTED_TEST_EXISTING_RESPONSE_FIELD = 'testExistingConnection'
EXPECTED_TEST_TABLEAU_RESPONSE_FIELD = 'testTableauCredentials'
EXPECTED_ADD_TABLEAU_RESPONSE_FIELD = 'addTableauAccount'
EXPECTED_LOOKER_METADATA_RESPONSE_FIELD = 'testLookerCredentials'
EXPECTED_LOOKER_GIT_SSH_RESPONSE_FILED = 'testLookerGitSshCredentials'
EXPECTED_LOOKER_GIT_CLONE_RESPONSE_FIELD = 'testLookerGitCloneCredentials'
EXPECTED_ADD_BI_RESPONSE_FIELD = 'addBiConnection'

# Available connections types
HIVE_MYSQL_CONNECTION_TYPE = 'hive-mysql'
HIVE_S3_CONNECTION_TYPE = 'hive-s3'
HIVE_SQL_CONNECTION_TYPE = 'hive'
PRESTO_SQL_CONNECTION_TYPE = 'presto'
PRESTO_S3_CONNECTION_TYPE = 'presto-s3'
GLUE_CONNECTION_TYPE = 'glue'
ATHENA_CONNECTION_TYPE = 'athena'
SPARK_CONNECTION_TYPE = 'spark'
REDSHIFT_CONNECTION_TYPE = 'redshift'
SNOWFLAKE_CONNECTION_TYPE = 'snowflake'
BQ_CONNECTION_TYPE = 'bigquery'
LOOKER_MD_CONNECTION_TYPE = 'looker'
LOOKER_GIT_SSH_CONNECTION_TYPE = 'looker-git-ssh'
LOOKER_GIT_CLONE_CONNECTION_TYPE = 'looker-git-clone'

# Available warehouse types
DATA_LAKE_WAREHOUSE_TYPE = 'data-lake'
REDSHIFT_WAREHOUSE_TYPE = REDSHIFT_CONNECTION_TYPE
SNOWFLAKE_WAREHOUSE_TYPE = SNOWFLAKE_CONNECTION_TYPE
BQ_WAREHOUSE_TYPE = BQ_CONNECTION_TYPE

CONNECTION_TO_WAREHOUSE_TYPE_MAP = {
    BQ_CONNECTION_TYPE: BQ_WAREHOUSE_TYPE,
    REDSHIFT_CONNECTION_TYPE: REDSHIFT_WAREHOUSE_TYPE,
    SNOWFLAKE_CONNECTION_TYPE: SNOWFLAKE_WAREHOUSE_TYPE,
}

# Available credential self-hosting mechanisms
SECRETS_MANAGER_CREDENTIAL_MECHANISM = 'secretsmanager'
SELF_HOSTING_MECHANISMS = [SECRETS_MANAGER_CREDENTIAL_MECHANISM]

# S3 event types
S3_METADATA_EVENT_TYPE = 's3_metadata_events'
S3_QL_EVENT_TYPE = 's3_ql_events'

# Job types
QL_JOB_TYPE = ['query_logs']

# Job limits
PRESTO_CATALOG_KEY = 'catalog_name'
HIVE_GET_PARTS_KEY = 'get_partition_locations'
HIVE_MAX_PARTS_KEY = 'max_partition_locations'
HIVE_MAX_PARTS_DEFAULT_VALUE = 50

# Certificate details
S3_CERT_MECHANISM = 'dc-s3'
PRESTO_CERT_PREFIX = 'certificates/presto/'
AWS_RDS_CA_CERT = 'https://s3.amazonaws.com/rds-downloads/rds-combined-ca-bundle.pem'

# Connections to friendly name (i.e. human presentable) map
GQL_TO_FRIENDLY_CONNECTION_MAP = {
    HIVE_MYSQL_CONNECTION_TYPE: 'Hive (metastore)',
    HIVE_S3_CONNECTION_TYPE: 'Hive (EMR logs)',
    HIVE_SQL_CONNECTION_TYPE: 'Hive (SQL)',
    PRESTO_SQL_CONNECTION_TYPE: PRESTO_SQL_CONNECTION_TYPE.capitalize(),
    PRESTO_S3_CONNECTION_TYPE: 'Presto (logs)',
    GLUE_CONNECTION_TYPE: GLUE_CONNECTION_TYPE.capitalize(),
    ATHENA_CONNECTION_TYPE: 'Athena',
    SPARK_CONNECTION_TYPE: 'Spark (SQL)',
    REDSHIFT_CONNECTION_TYPE: REDSHIFT_CONNECTION_TYPE.capitalize(),
    SNOWFLAKE_CONNECTION_TYPE: SNOWFLAKE_CONNECTION_TYPE.capitalize(),
    BQ_CONNECTION_TYPE: 'BigQuery',
    LOOKER_MD_CONNECTION_TYPE: 'Looker (metadata)',
    LOOKER_GIT_SSH_CONNECTION_TYPE: 'Looker ML (git)',
    LOOKER_GIT_CLONE_CONNECTION_TYPE: 'Looker ML (git)',
}

# Verbiage
CONNECTION_TEST_SUCCESS_VERBIAGE = 'Connection test was successful!'
CONNECTION_TEST_FAILED_VERBIAGE = 'Connection test failed!'
ADD_CONNECTION_SUCCESS_VERBIAGE = 'Success! Added connection for '
ADD_CONNECTION_FAILED_VERBIAGE = 'Failed to add connection!'
VALIDATIONS_FAILED_VERBIAGE = 'Some validations failed. Would you like to create the connection anyway?'
OPERATION_ERROR_VERBIAGE = 'Operation failed - This might not be a valid connection for your account. ' \
                           'Please contact Monte Carlo.'
CONFIRM_CONNECTION_VERBIAGE = 'Please confirm you want to add this connection'
SKIP_ADD_CONNECTION_VERBIAGE = 'Skipping adding the connection.'



"""
Common commands
"""
import click

from montecarlodata.tools import convert_uuid_callback

RESOURCE_VERBIAGE = 'This can be helpful if the resource and collector are in different accounts'

DISAMBIGUATE_DC_OPTIONS = [
    click.option('--collector-id', 'dc_id', required=False, type=click.UUID, callback=convert_uuid_callback,
                 help='ID for the data collector. To disambiguate accounts with multiple collectors.'),
]

DC_RESOURCE_OPTIONS = [
    click.option('--resource-aws-region', required=False,
                 help='Override the AWS region where the resource is located. '
                      'Defaults to the region where the collector is hosted.'),
    click.option('--resource-aws-profile', required=False,
                 help=f'Override the AWS profile use by the CLI for the resource. {RESOURCE_VERBIAGE}.')
]


from typing import Dict, Optional, List

import click
import click_config_file

from montecarlodata import settings
from montecarlodata.common.commands import DISAMBIGUATE_DC_OPTIONS, DC_RESOURCE_OPTIONS
from montecarlodata.common.resources import CloudResourceService
from montecarlodata.discovery.networking import NetworkDiscoveryService
from montecarlodata.discovery.policy_gen import PolicyDiscoveryService
from montecarlodata.tools import add_common_options

# Shared command verbiage
PROFILE_VERBIAGE = 'If not specified, the one in the Monte Carlo CLI profile is used'
RESOURCE_VERBIAGE = 'This can be helpful if the resource and collector are in different accounts'
DISCOVERY_REVIEW_VERBIAGE = 'After review, output of this command can be redirected into ' \
                            '`montecarlo integrations create-role` or `montecarlo discovery cf-role-gen` ' \
                            'if you prefer IaC'

GLUE_COMMON_OPTIONS = [
    click.option('--database-name', 'database_names', required=True, multiple=True,
                 help='Glue/Athena database name to generate a policy from. Enter \'\\*\' to give Monte Carlo access '
                      'to all databases. This option can be passed multiple times for more than one database.'),
    click.option('--data-bucket-name', 'bucket_names', required=False, multiple=True,
                 help='Name of a S3 bucket storing the data for your Glue/Athena tables. If this option is not '
                      'specified the bucket names are derived (looked up) from the tables in your databases. This '
                      'option can be passed multiple times for more than one bucket. Enter \'\\*\' to give Monte '
                      'Carlo access to all buckets.')
]


@click.group(help='Display information about resources.')
def discovery():
    """
    Group for any discovery related subcommands
    """
    pass


@discovery.command(help='List details about EMR clusters in a region.')
@click.option('--aws-profile', help=f'AWS profile. {PROFILE_VERBIAGE}', required=False)
@click.option('--aws-region', help=f'AWS region. {PROFILE_VERBIAGE}', required=False)
@click.option('--only-log-locations', help='Display only unique log locations', is_flag=True, default=False)
@click.option('--created-after', help='Display clusters created after date (e.g. 2017-07-04T00:01:30)', required=False)
@click.option('--state', help='Cluster states', required=False, type=click.Choice(['active', 'terminated', 'failed']),
              multiple=True)
@click.option('--no-grid',
              help='Do not display as grid and print as results are available, useful when the cluster list is large',
              is_flag=True, default=False)
@click_config_file.configuration_option(settings.OPTION_FILE_FLAG)
@click.pass_obj
def list_emr_clusters(ctx: Dict, aws_profile: Optional[str] = None, aws_region: Optional[str] = None,
                      only_log_locations: Optional[bool] = False, created_after: Optional[str] = None,
                      state: Optional[List] = None, no_grid: Optional[bool] = False) -> None:
    CloudResourceService(config=ctx['config'], aws_profile_override=aws_profile, aws_region_override=aws_region) \
        .list_emr_clusters(only_log_locations=only_log_locations, created_after=created_after, states=state,
                           no_grid=no_grid)


@discovery.command(help='Alpha network recommender. Attempts to analyze and makes recommendations on how to connect a '
                        'resource with the Data Collector.')
@click.pass_obj
@click.option('--resource-identifier', required=True,
              help='Identifier for the AWS resource you want to connect the Collector with (e.g. Redshift cluster ID).')
@click.option('--resource-type', required=True, help='Type of AWS resource.',
              type=click.Choice(NetworkDiscoveryService.MCD_NETWORK_REC_RESOURCE_TYPE_MAP.keys()))
@click.option('--collector-aws-profile', required=False,
              help=f'Override the AWS profile use by the CLI for the Collector. {RESOURCE_VERBIAGE}.')
@add_common_options(DC_RESOURCE_OPTIONS)
@add_common_options(DISAMBIGUATE_DC_OPTIONS)
@click_config_file.configuration_option(settings.OPTION_FILE_FLAG)
def network_recommender(ctx, **kwargs):
    NetworkDiscoveryService(config=ctx['config'], aws_wrapper=None).recommend_network_dispatcher(**kwargs)


@discovery.command(help='Generate a CloudFormation template to create a resource access IAM role. '
                        'After review, this template can be deployed using CloudFormation. '
                        'The Role ARN and External ID for onboarding can be found in the stack outputs.')
@click.pass_obj
@click.option('--policy-file', 'policy_files', required=True, multiple=True, type=click.File(),
              help='File containing an IAM policy to generate an IAM role from. This option can be passed '
                   'multiple times for more than one policy.')
@add_common_options(DISAMBIGUATE_DC_OPTIONS)
def cf_role_gen(ctx, **kwargs):
    PolicyDiscoveryService(config=ctx['config'], aws_wrapper=None).generate_cf_role(**kwargs)


@discovery.command(help=f'Generate an IAM policy for Glue. {DISCOVERY_REVIEW_VERBIAGE}.')
@click.pass_obj
@add_common_options(GLUE_COMMON_OPTIONS)
@add_common_options(DC_RESOURCE_OPTIONS)
@add_common_options(DISAMBIGUATE_DC_OPTIONS)
def glue_policy_gen(ctx, **kwargs):
    PolicyDiscoveryService(config=ctx['config'], aws_wrapper=None).generate_glue_policy(**kwargs)


@discovery.command(help=f'Generate an IAM policy for Athena. {DISCOVERY_REVIEW_VERBIAGE}.')
@click.pass_obj
@add_common_options(GLUE_COMMON_OPTIONS)
@click.option('--workgroup-name', required=True, default='primary', show_default=True,
              help='Athena workgroup for Monte Carlo to use when performing queries. '
                   'The "primary" workgroup for the region is used if one is not specified.')
@add_common_options(DC_RESOURCE_OPTIONS)
@add_common_options(DISAMBIGUATE_DC_OPTIONS)
def athena_policy_gen(ctx, **kwargs):
    PolicyDiscoveryService(config=ctx['config'], aws_wrapper=None).generate_athena_policy(**kwargs)



from montecarlodata.config import Config
from montecarlodata.errors import manage_errors
from montecarlodata.integrations.onboarding.base import BaseOnboardingService
from montecarlodata.integrations.onboarding.fields import EXPECTED_GLUE_GQL_RESPONSE_FIELD, \
    EXPECTED_ATHENA_GQL_RESPONSE_FIELD, GLUE_CONNECTION_TYPE, ATHENA_CONNECTION_TYPE
from montecarlodata.queries.onboarding import TEST_GLUE_CRED_MUTATION, TEST_ATHENA_CRED_MUTATION


class GlueAthenaOnboardingService(BaseOnboardingService):

    def __init__(self, config: Config, **kwargs):
        super().__init__(config, **kwargs)

    @manage_errors
    def onboard_glue(self, **kwargs) -> None:
        """
        Onboard a glue connection by validating and adding a connection
        """
        self.onboard(validation_query=TEST_GLUE_CRED_MUTATION,
                     validation_response=EXPECTED_GLUE_GQL_RESPONSE_FIELD,
                     connection_type=GLUE_CONNECTION_TYPE, **kwargs)

    @manage_errors
    def onboard_athena(self, **kwargs) -> None:
        """
        Onboard an athena connection by validating and adding a connection
        """
        self.onboard(validation_query=TEST_ATHENA_CRED_MUTATION,
                     validation_response=EXPECTED_ATHENA_GQL_RESPONSE_FIELD,
                     connection_type=ATHENA_CONNECTION_TYPE, **kwargs)



import json
from typing import Optional, Dict
from uuid import UUID

import click

from montecarlodata.common.data import MonolithResponse
from montecarlodata.config import Config
from montecarlodata.errors import manage_errors
from montecarlodata.integrations.onboarding.base import BaseOnboardingService
from montecarlodata.integrations.onboarding.fields import EXPECTED_UPDATE_CREDENTIALS_RESPONSE_FIELD, \
    EXPECTED_REMOVE_CONNECTION_RESPONSE_FIELD, OPERATION_ERROR_VERBIAGE, EXPECTED_TEST_EXISTING_RESPONSE_FIELD
from montecarlodata.queries.onboarding import UPDATE_CREDENTIALS_MUTATION, REMOVE_CONNECTION_MUTATION, \
    TEST_EXISTING_CONNECTION_QUERY


class ConnectionOperationsService(BaseOnboardingService):
    REMOVE_PROMPT = 'Are you sure? This action deletes any associated jobs, monitors, etc.'

    def __init__(self, config: Config, **kwargs):
        super().__init__(config, **kwargs)

    @manage_errors
    def update_credentials(
        self,
        connection_id: UUID,
        changes: Dict,
        should_validate: Optional[bool] = True,
        should_replace: Optional[bool] = False
    ) -> None:
        """
        Update credentials for a connection
        """
        self.echo_operation_status(
            self._request_wrapper.make_request_v2(
                query=UPDATE_CREDENTIALS_MUTATION,
                operation=EXPECTED_UPDATE_CREDENTIALS_RESPONSE_FIELD,
                variables=dict(connection_id=str(connection_id), changes=json.dumps(changes),
                               should_validate=should_validate, should_replace=should_replace)
            ),
            operation=f'Updated \'{connection_id}\'.'
        )

    @manage_errors
    def remove_connection(self, connection_id: UUID, no_prompt: Optional[bool] = False) -> None:
        """
        Remove connection by ID. Deletes any associated jobs
        """
        if no_prompt or click.confirm(self.REMOVE_PROMPT, abort=True):
            self.echo_operation_status(
                self._request_wrapper.make_request_v2(
                    query=REMOVE_CONNECTION_MUTATION,
                    operation=EXPECTED_REMOVE_CONNECTION_RESPONSE_FIELD,
                    variables=dict(connection_id=str(connection_id))
                ),
                operation=f'Removed \'{connection_id}\'.'
            )

    @manage_errors
    def echo_test_existing(self, connection_id: UUID) -> None:
        """
        Tests an existing connection and echos results in pretty JSON
        """
        response = self._request_wrapper.make_request_v2(
            query=TEST_EXISTING_CONNECTION_QUERY,
            operation=EXPECTED_TEST_EXISTING_RESPONSE_FIELD,
            variables=dict(connection_id=connection_id)
        )
        click.echo(json.dumps(response.data, indent=4, ensure_ascii=False))

    @staticmethod
    def echo_operation_status(response: MonolithResponse, operation: str) -> None:
        """
        Echos operation status. Expects response contains a 'success' field as per API signature
        """
        if response.data.get('success'):
            click.echo(f'Success! {operation}')
        else:
            # Any errors thrown by the API are captured and handled in the request
            # lib/caller deco (e.g. validation errors). This should only reachable for invalid inputs for the account
            click.echo(OPERATION_ERROR_VERBIAGE)



import click
import click_config_file

import montecarlodata.settings as settings
from montecarlodata.collector.network_tests import CollectorNetworkTestService
from montecarlodata.collector.fields import DEFAULT_COLLECTION_REGION
from montecarlodata.collector.management import CollectorManagementService
from montecarlodata.tools import validate_json_callback, add_common_options
from montecarlodata.common.commands import DISAMBIGUATE_DC_OPTIONS

# Shared command verbiage
PROFILE_VERBIAGE = 'If not specified, the one in the Monte Carlo CLI profile is used'
ADD_DC_VERBIAGE = 'Prompts to opens browser to CF console. If declined (or skipped) can use deploy, get-template ' \
                  'or open-link with the generated ID.'

# Options shared across commands
REGION_OPTIONS = [
    *DISAMBIGUATE_DC_OPTIONS,
    click.option('--aws-region', required=False, default=DEFAULT_COLLECTION_REGION, show_default=True,
                 help='AWS region where the collector is deployed or intended to be deployed.')
]
DEPLOYMENT_OPTIONS = [
    click.option('--aws-profile', required=False, help=f'AWS profile. {PROFILE_VERBIAGE}.'),
    click.option('--params', required=False, default=None, callback=validate_json_callback,
                 help="""
                  Parameters key,value pairs as JSON. If a key is not specified the existing (or default) value is used.
                  \b
                  \n
                  E.g. --params '{"CreateEventInfra":"True"}'
                  """),  # \b disables wrapping
    *REGION_OPTIONS
]

NETWORK_TEST_OPTIONS = [
    *DISAMBIGUATE_DC_OPTIONS,
    click.option('--host', required=True, help='Host to check.'),
    click.option('--port', required=True, type=click.INT, help='Port to check.'),
    click.option('--timeout', required=False, default=5, type=click.INT, show_default=True, help='Timeout in seconds.')
]


@click.group(help='Manage a data collector.')
def collectors():
    """
    Group for any collector related subcommands
    """
    pass


@collectors.command(help='List all collector records.', name='list')
@click.pass_obj
@click.option('--active-only', required=False, default=False, show_default=True, is_flag=True,
              help='Only list active collectors.')
@click_config_file.configuration_option(settings.OPTION_FILE_FLAG)
def list_collectors(ctx, active_only):
    CollectorManagementService(config=ctx['config'], aws_wrapper=None).echo_collectors(active_only=active_only)


@collectors.command(help=f'Add a collector record to the account. {ADD_DC_VERBIAGE}', name='add')
@click.pass_obj
@click.option('--no-prompt', required=False, default=False, show_default=True, is_flag=True,
              help='Skip prompt for launching browser (Auto no).')
@click_config_file.configuration_option(settings.OPTION_FILE_FLAG)
def add_collector(ctx, no_prompt):
    CollectorManagementService(config=ctx['config']).add_collector(no_prompt=no_prompt)


@collectors.command(help='Get link to the latest template. For initial deployment or manually upgrading.')
@click.pass_obj
@add_common_options(REGION_OPTIONS)
@click_config_file.configuration_option(settings.OPTION_FILE_FLAG)
def get_template(ctx, aws_region, **kwargs):
    """
    Get the latest template for this account
    """
    CollectorManagementService(config=ctx['config'], aws_region_override=aws_region).echo_template(**kwargs)


@collectors.command(help='Opens browser to CF console with a quick create link. For initial deployment.')
@click.pass_obj
@click.option('--dry', required=False, default=False, show_default=True, is_flag=True, help='Echos quick create link.')
@add_common_options(REGION_OPTIONS)
@click_config_file.configuration_option(settings.OPTION_FILE_FLAG)
def open_link(ctx, aws_region, **kwargs):
    """
    Open browser with a quick create link for deploying a data collector
    """
    CollectorManagementService(config=ctx['config'], aws_region_override=aws_region).launch_quick_create_link(**kwargs)


@collectors.command(help='Deploy a data collector stack.')
@click.pass_obj
@add_common_options(DEPLOYMENT_OPTIONS)
@click.option('--stack-name', required=True, help='The name that is associated with the CloudFormation stack. '
                                                  'Must be unique in the region.')
@click.option('--enable-termination-protection/--no-enable-termination-protection', 'termination_protection',
              default=False, show_default=True, help='Enable termination protection for this stack.')
@click_config_file.configuration_option(settings.OPTION_FILE_FLAG)
def deploy(ctx, aws_profile, aws_region, params, **kwargs):
    """
    Deploy a collector for this account
    """
    CollectorManagementService(config=ctx['config'], aws_profile_override=aws_profile,
                               aws_region_override=aws_region).deploy_template(new_params=params, **kwargs)


@collectors.command(help='Upgrade to the latest version.')
@click.pass_obj
@add_common_options(DEPLOYMENT_OPTIONS)
@click_config_file.configuration_option(settings.OPTION_FILE_FLAG)
def upgrade(ctx, aws_profile, aws_region, params, **kwargs):
    """
    Upgrade the collector for this account
    """
    CollectorManagementService(config=ctx['config'], aws_profile_override=aws_profile,
                               aws_region_override=aws_region).upgrade_template(new_params=params, **kwargs)


@collectors.command(help='Checks if telnet connection is usable from the collector.')
@click.pass_obj
@add_common_options(NETWORK_TEST_OPTIONS)
@click_config_file.configuration_option(settings.OPTION_FILE_FLAG)
def test_telnet(ctx, **kwargs):
    """
    Network debugging utility to test telnet via the collector
    """
    CollectorNetworkTestService(config=ctx['config']).echo_telnet_test(**kwargs)


@collectors.command(help='Tests if a destination exists and accepts requests. '
                        'Opens a TCP Socket to a specific port from the collector.')
@click.pass_obj
@add_common_options(NETWORK_TEST_OPTIONS)
@click_config_file.configuration_option(settings.OPTION_FILE_FLAG)
def test_tcp_open(ctx, **kwargs):
    """
    Network debugging utility to test TCP open via the collector
    """
    CollectorNetworkTestService(config=ctx['config']).echo_tcp_open_test(**kwargs)



import configparser
import os
from dataclasses import dataclass
from typing import Optional

import click

import montecarlodata.settings as settings
from montecarlodata.fs_utils import mkdirs


@dataclass
class Config:
    mcd_id: str
    mcd_token: str
    aws_profile: Optional[str] = None
    aws_region: Optional[str] = None


class ConfigManager:
    def __init__(self, profile_name: str, base_path: str, config_parser: Optional[configparser.ConfigParser] = None):
        self._profile_name = profile_name
        self._base_path = base_path
        self._profile_config_file = os.path.join(self._base_path, settings.PROFILE_FILE_NAME)

        self._config = config_parser or configparser.ConfigParser()
        self._config.read(self._profile_config_file)

    def write(self, **kwargs) -> None:
        """
        Write any configuration key value pairs to the specified section (profile name)
        """
        if self._profile_name not in self._config.sections():
            # if the section does not exist add it
            self._config.add_section(self._profile_name)

        for k, v in kwargs.items():
            self._config.set(self._profile_name, k, v)

        mkdirs(self._base_path)
        with open(self._profile_config_file, 'w') as cf:
            self._config.write(cf)

    def read(self) -> Optional[Config]:
        """
        Return configuration from section (profile name) if it exists.
        Any MCD values can be overwritten by the environment. Uses system default for AWS if not set.
        """
        try:
            return Config(
                mcd_id=settings.MCD_DEFAULT_API_ID or self._config.get(self._profile_name, 'mcd_id'),
                mcd_token=settings.MCD_DEFAULT_API_TOKEN or self._config.get(self._profile_name, 'mcd_token'),
                aws_profile=self._config.get(self._profile_name, 'aws_profile', fallback=None),
                aws_region=self._config.get(self._profile_name, 'aws_region', fallback=None)
            )
        except configparser.NoSectionError:
            click.echo(f"Failed to find configuration for '{self._profile_name}'. Please setup using 'configure' first")


# Query responses
EXPECTED_GET_INSIGHTS_FIELD = 'getInsights'
EXPECTED_GET_REPORT_FIELD = 'getReportUrl'

# Insight Names
KEY_ASSETS_INSIGHT_NAME = 'key_assets'
COVERAGE_OVERVIEW_INSIGHT_NAME = 'monitors'
CLEANUP_SUGGESTIONS_INSIGHT_NAME = 'cleanup_suggestions'
EVENTS_INSIGHT_NAME = 'events'
READ_WRITE_INSIGHT_NAME = 'table_read_write_stats'
INCIDENT_QUERY_INSIGHT_NAME = 'incident_query_changes'
DETERIORATING_QUERY_INSIGHT_NAME = 'query_runtime_trends'
RULES_AND_SLIS_INSIGHT_NAME = 'rule_and_sli_results'

# Supported Schemes
FILE_SCHEME = 'file'
S3_SCHEME = 's3'
SCHEME_DELIM = '://'

# Insights extension
INSIGHTS_DEFAULT_EXTENSION = '.csv'

# Insights headers
LIST_INSIGHTS_HEADERS = ['Insight', 'Description', 'Available']



import click

from montecarlodata.common.common import read_as_base64
from montecarlodata.config import Config
from montecarlodata.errors import manage_errors, prompt_connection, complain_and_abort
from montecarlodata.integrations.onboarding.base import BaseOnboardingService
from montecarlodata.integrations.onboarding.fields import EXPECTED_TEST_TABLEAU_RESPONSE_FIELD, \
    CONFIRM_CONNECTION_VERBIAGE, CONNECTION_TEST_SUCCESS_VERBIAGE, CONNECTION_TEST_FAILED_VERBIAGE, \
    SKIP_ADD_CONNECTION_VERBIAGE, EXPECTED_ADD_TABLEAU_RESPONSE_FIELD, ADD_CONNECTION_SUCCESS_VERBIAGE, \
    ADD_CONNECTION_FAILED_VERBIAGE, EXPECTED_LOOKER_METADATA_RESPONSE_FIELD, LOOKER_MD_CONNECTION_TYPE, \
    EXPECTED_ADD_BI_RESPONSE_FIELD, EXPECTED_LOOKER_GIT_CLONE_RESPONSE_FIELD, LOOKER_GIT_CLONE_CONNECTION_TYPE
from montecarlodata.queries.onboarding import TEST_TABLEAU_CRED_MUTATION, ADD_TABLEAU_CONNECTION_MUTATION, \
    TEST_LOOKER_METADATA_CRED_MUTATION, ADD_BI_CONNECTION_MUTATION, TEST_LOOKER_GIT_CLONE_CRED_MUTATION


class ReportsOnboardingService(BaseOnboardingService):
    def __init__(self, config: Config, **kwargs):
        super().__init__(config, **kwargs)

    @manage_errors
    def onboard_tableau(self, **kwargs) -> None:
        """
        Onboard a tableau connection
        """
        connection_options = self._build_connection_options(**kwargs)
        response = self._request_wrapper.make_request_v2(
            query=TEST_TABLEAU_CRED_MUTATION,
            operation=EXPECTED_TEST_TABLEAU_RESPONSE_FIELD,
            variables=connection_options.monolith_base_payload
        )
        if response.data.success:
            click.echo(CONNECTION_TEST_SUCCESS_VERBIAGE)
        else:
            complain_and_abort(CONNECTION_TEST_FAILED_VERBIAGE)

        if not connection_options.validate_only:
            if connection_options.dc_id:
                connection_options.monolith_base_payload['dc_id'] = connection_options.dc_id

            prompt_connection(message=CONFIRM_CONNECTION_VERBIAGE, skip_prompt=connection_options.auto_yes)
            response = self._request_wrapper.make_request_v2(
                query=ADD_TABLEAU_CONNECTION_MUTATION,
                operation=EXPECTED_ADD_TABLEAU_RESPONSE_FIELD,
                variables=connection_options.monolith_base_payload
            )
            if response.data.tableauAccount.uuid:
                click.echo(f'{ADD_CONNECTION_SUCCESS_VERBIAGE}tableau.')
            else:
                complain_and_abort(ADD_CONNECTION_FAILED_VERBIAGE)
        else:
            click.echo(SKIP_ADD_CONNECTION_VERBIAGE)

    @manage_errors
    def onboard_looker_metadata(self, **kwargs) -> None:
        """
        Onboard a looker metadata connection
        """
        self.onboard(validation_query=TEST_LOOKER_METADATA_CRED_MUTATION,
                     validation_response=EXPECTED_LOOKER_METADATA_RESPONSE_FIELD,
                     connection_query=ADD_BI_CONNECTION_MUTATION,
                     connection_response=EXPECTED_ADD_BI_RESPONSE_FIELD,
                     connection_type=LOOKER_MD_CONNECTION_TYPE, **kwargs)

    @manage_errors
    def onboard_looker_git(self, **kwargs) -> None:
        """
        Onboard a looker git ssh connection
        """
        if kwargs.get('ssh_key'):
            kwargs['ssh_key'] = read_as_base64(kwargs.pop('ssh_key')).decode('utf-8')
        self.onboard(validation_query=TEST_LOOKER_GIT_CLONE_CRED_MUTATION,
                     validation_response=EXPECTED_LOOKER_GIT_CLONE_RESPONSE_FIELD,
                     connection_query=ADD_BI_CONNECTION_MUTATION,
                     connection_response=EXPECTED_ADD_BI_RESPONSE_FIELD,
                     connection_type=LOOKER_GIT_CLONE_CONNECTION_TYPE, **kwargs)
